use core::future::*;
use core::pin::Pin;
use core::task::*;

pub struct DontAbort<A> {
    a: A,
    done: bool
}

pub struct DontAbortFuture<'a, A> {
    inner: &'a mut DontAbort<A>
}

impl<A> DontAbort<A> {
    pub fn new(a: A) -> DontAbort<A> {
        DontAbort { a, done: false }
    }

    #[allow(dead_code)]
    pub fn done(self: &Self) -> bool {
        return self.done;
    }

    pub fn continue_wait<'a>(self: &'a mut Self) -> DontAbortFuture<'a, A> {
        DontAbortFuture { inner: self }
    }
}

impl<A: Unpin> Unpin for DontAbort<A> {}
impl<'a, A: Unpin> Unpin for DontAbortFuture<'a, A> {}

impl<'a, A> Future for DontAbortFuture<'a, A>
where
    A: Future,
{
    type Output = A::Output;

    fn poll(self: Pin<&mut Self>, cx: &mut Context<'_>) -> Poll<Self::Output> {
        let this = unsafe { self.get_unchecked_mut() };
        if this.inner.done {
            panic!("Ready future is called again!");
        }
        let a = unsafe { Pin::new_unchecked(&mut this.inner.a) };
        if let Poll::Ready(x) = a.poll(cx) {
            this.inner.done = true;
            return Poll::Ready(x);
        }
        Poll::Pending
    }
}
